;;; GRDOC.EL --- grdoc routines for Emacs -*-byte-compile-dynamic-docstrings: nil;-*-

;; Author: Paul Walker <pwalker@ncsa.uiuc.edu>
;; Maintainer: Paul Walker <pwalker@ncsa.uiuc.edu>
;; Created: Wed Jan 17 1996
;; Version: $Revision: 1.19 $
;; Keywords: languages c hypermedia docs local tools

;; UNIVERSITY OF ILLINOIS (UI), NATIONAL CENTER FOR SUPERCOMPUTING
;; APPLICATIONS (NCSA), Software Distribution Policy for Copyrighted
;; Software

;; The NCSA software "grdoc" Source code and associated scripts is
;; copyrighted, but available without fee for education, academic
;; research and non-commercial purposes. The software is copyrighted
;; in the name of the UI, and ownership of the software remains with
;; the UI.  Users may distribute the binary and source code to third
;; parties provided that the copyright notice and this statement
;; appears on all copies and that no charge is made for such
;; copies. Any entity wishing to integrate all or part of the source
;; code into a product for commercial use or resale, should contact
;; the University of Illinois, c/o NCSA, to negotiate an appropriate
;; license for such commercial use.

;; UI MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THE SOFTWARE
;; FOR ANY PURPOSE. IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED
;; WARRANTY. THE UI SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY
;; THE USER OF THIS SOFTWARE. The software may have been developed
;; under agreements between the UI and the Federal Government which
;; entitle the Government to certain rights.

;; By copying this program, you, the user, agree to abide by the
;; copyright conditions and understandings with respect to any
;; software which is marked with a copyright notice.

;;; LCD Archive Entry:
;;; grdoc-2|Paul Walker|pwalker@ncsa.uiuc.edu
;;; |grdoc routines for Emacs
;;; |$Date: 1996/01/28 05:39:36 $|$Revision: 1.19 $|

;;; Commentary:

;; grdoc is a commenting system developed at the NCSA/WashU GR groups
;; by Paul Walker with Joan Masso and Alan Shutko.  It is documented
;; at http://jean-luc.ncsa.uiuc.edu/Codes/grdoc/ .  grdoc.el is a
;; minor mode intended to make it easier to format comments for use
;; with the grdoc formatting system.  It was originally developed by
;; Pauil Walker <pwalker@ncsa.uiuc.edu> and extended by Alan Shutko
;; <ats@hubert.wustl.edu>.

;; To use, place the following in your .emacs:

;; (require 'grdoc)
;; (add-hook 'fortran-mode-hook 'turn-on-grdoc-mode)
;; (add-hook 'c-mode-hook 'turn-on-grdoc-mode) 
;; and similar for any other language modes for which you would like to use grdoc.

;; A few things to note;

;; 1. This needs tempo.  tempo.el is included in recent versions of
;;    Emacs and Xemacs.
;; 2. To avoid interactivity, do
;;    (setq grdoc-use-interactive nil)
;; 3. To force c-style comments in fortran et al do
;;    (setq-default grdoc-comment-start " ")
;; 4. To turn off AI, do
;;    (grdoc-use-ai nil)


;;; Code:
(require 'tempo)
(require 'cl)
(require 'easymenu)

(defconst grdoc-version (substring "$Revision: 1.19 $" 11 -2)
  "$Id: grdoc.el,v 1.19 1996/01/28 05:39:36 ats Exp $

Report bugs to: Alan Shutko <ats@hubert.wustl.edu>")

(defun grdoc-version () (interactive)
  (message "grdoc version %s" grdoc-version))

(defvar grdoc-use-ai t
  "*If non-nil, parse source files to fill in templates.")

(defvar grdoc-mode-map (make-sparse-keymap))
(define-key grdoc-mode-map "\C-cgf"      'grdoc-file-header)
(define-key grdoc-mode-map "\C-cge"      'grdoc-header-header)
(define-key grdoc-mode-map "\C-cgr"      'grdoc-routine-header)
(define-key grdoc-mode-map "\C-cgv"      'grdoc-var-element)
(define-key grdoc-mode-map "\C-cgp"      'grdoc-par-element)
(define-key grdoc-mode-map "\C-cgh"      'grdoc-history-element)

(easy-menu-define grdoc-menu-map 
		  grdoc-mode-map
		  "GR Doc menus"
		  (list "grdoc"
			["Insert file header" grdoc-file-header t]
			["Insert header comment" grdoc-header-header t]
			["Insert routine header" grdoc-routine-header t]
			["Insert variable comment" grdoc-var-element t]
			["Insert parameter comment" grdoc-par-element t]
			["Insert history comment" grdoc-history-element t]
			(if grdoc-use-ai
			    ["Turn off AI" grdoc-use-ai t]
			  ["Turn on AI" grdoc-use-ai t])
			["Set prepend char" grdoc-set-prepend-char t]))

(defvar grdoc-mode nil)
(make-variable-buffer-local 'grdoc-mode)

(defun grdoc-use-ai (&optional arg)
  "Toggle use of AI in grdoc mode
With arg, turn on AI iff arg is positive.  The AI support in grdoc-mode 
will scan the source for items to fill in templates, such as routine names, 
variable types, etc."
  (interactive "P")
  (let ((on-p (if arg (> (prefix-numeric-value arg) 0) (not grdoc-use-ai))))
    (set 'grdoc-use-ai on-p)
    (if on-p
	(progn
	  (message "grdoc will use AI")
	  (easy-menu-change nil "grdoc" 		  
			    (list "grdoc"
				  ["Insert file header" grdoc-file-header t]
				  ["Insert header comment" grdoc-header-header t]
				  ["Insert routine header" grdoc-routine-header t]
				  ["Insert variable comment" grdoc-var-element t]
				  ["Insert parameter comment" grdoc-par-element t]
				  ["Insert history comment" grdoc-history-element t]
				  ["Turn off AI" grdoc-use-ai t]
				  ["Set prepend char" grdoc-set-prepend-char t])))
      (message "grdoc will not use AI")
      (easy-menu-change nil "grdoc" 		  
			(list "grdoc"
			      ["Insert file header" grdoc-file-header t]
			      ["Insert header comment" grdoc-header-header t]
			      ["Insert routine header" grdoc-routine-header t]
			      ["Insert variable comment" grdoc-var-element t]
			      ["Insert parameter comment" grdoc-par-element t]
			      ["Insert history comment" grdoc-history-element t]
			      ["Turn on AI" grdoc-use-ai t]
			      ["Set prepend char" grdoc-set-prepend-char t])))))

(defun grdoc-mode (&optional arg)
  "Toggle grdoc-mode.
With arg, turn grdoc-mode on iff arg is positive.  grdoc-mode 
makes it easier to create comments which work with the grdoc 
commenting system.  Runs `grdoc-mode-hook' on startup.  It binds
the following keys:

\\{grdoc-mode-map}"
  (interactive "P")
  (let ((on-p (if arg (> (prefix-numeric-value arg) 0) (not grdoc-mode))))
    (set 'grdoc-mode on-p)
    (if on-p
	(progn
	  (grdoc-set-comment-start)
	  (easy-menu-add grdoc-menu-map)
	  (run-hooks 'grdoc-mode-hooks))
      (easy-menu-remove grdoc-menu-map))))

(defun turn-on-grdoc-mode ()
  (grdoc-mode 1))

(defvar grdoc-mode-hook nil
  "Hooks run when starting grdoc-mode.")
  
(defvar grdoc-comment-start nil
  "*Character which defines the start of a grdoc comment region"
)
(make-variable-buffer-local 'grdoc-comment-start)

(defvar grdoc-use-interactive t
  "*Set to nil if you don't want to use interactive grdoc comments"
)

(defvar grdoc-ai-var-style 'multi
  "*Style of handling variable comments.
If 'multi, include multiple variables declared on the same line in the 
same @var comment.  If 'one, put each variable in its own comment.")

(make-variable-buffer-local 'grdoc-comment-start)

(defun grdoc-set-prepend-char (char)
  "Sets `grdoc-comment-start'."
  (interactive "sPrepend char:  ")
  (setq grdoc-comment-start char))

(defun grdoc-set-comment-start ()
  (interactive)
  (cond
   (grdoc-comment-start ())
   ((eq major-mode 'fortran-mode) (setq grdoc-comment-start "c"))
   ((eq major-mode 'perl-mode) (setq grdoc-comment-start "#"))
   ((eq major-mode 'idl-mode) (setq grdoc-comment-start "\;\;"))
   ((eq major-mode 'emacs-lisp-mode) (setq grdoc-comment-start "\;\;"))
   (t (setq grdoc-comment-start " "))
   )
)

(tempo-define-template 
 "grdoc-file-header"
  '(
    grdoc-comment-start
    "/*@@\n"
    grdoc-comment-start
    "  @file      "
    (file-name-nondirectory (buffer-file-name)) "\n"
    grdoc-comment-start
    "  @date      "
    (current-time-string) "\n"
    grdoc-comment-start
    "  @author    " 
    (user-full-name)
    'n
    grdoc-comment-start
    "  @desc \n"
    grdoc-comment-start
    "  "
    'p
    'n
    grdoc-comment-start
    "  @enddesc \n"
    grdoc-comment-start
    "@@*/\n"
    )
)

(tempo-define-template 
 "grdoc-header-header"
  '(
    grdoc-comment-start
    "/*@@\n"
    grdoc-comment-start
    "  @header    "
    (file-name-nondirectory (buffer-file-name)) "\n"
    grdoc-comment-start
    "  @date      "
    (current-time-string) "\n"
    grdoc-comment-start
    "  @author    " 
    (user-full-name)
    'n
    grdoc-comment-start
    "  @desc \n"
    grdoc-comment-start
    "  "
    'p
    'n
    grdoc-comment-start
    "  @enddesc \n"
    grdoc-comment-start
    "@@*/\n"
    )
)


(tempo-define-template 
 "grdoc-routine-header"
  '(
    grdoc-comment-start
    "/*@@\n"
    grdoc-comment-start
    "  @routine    "
;    '(p "Routine:  ")
    (grdoc-routine-name)
    'n
    grdoc-comment-start
    "  @date       "
    (current-time-string) "\n"
    grdoc-comment-start
    "  @author     " 
    (user-full-name)
    'n
    grdoc-comment-start
    "  @desc \n"
    grdoc-comment-start
    "  "
    'p
    "\n"
    grdoc-comment-start
    "  @enddesc \n"
    grdoc-comment-start
    "  @calls     " (grdoc-calls)
    grdoc-comment-start
    "  @calledby   \n"
    grdoc-comment-start
    "  @history \n"
    grdoc-comment-start
    "\n"
    grdoc-comment-start
    "  @endhistory \n"
    (grdoc-insert-variables)
    "@@*/\n"
    )
)

(tempo-define-template
 "grdoc-var-element"
  '(
    grdoc-comment-start
    "  @var     "
    '(p "Variable:  ")
    'n
    grdoc-comment-start
    "  @vdesc   "
    '(p "Description:  ")
    'n
    grdoc-comment-start
    "  @vtype   " 
    '(p "Type:  ")
    'n
    grdoc-comment-start
    "  @vio     "
;    '(p "io:  ")
    (grdoc-prompt-vio)
    'n
    grdoc-comment-start
    "  @vcomment \n"
    grdoc-comment-start
    'p 'n
    grdoc-comment-start
    "  @endvar \n"
    )
)

(tempo-define-template
 "grdoc-par-element"
  '(
    grdoc-comment-start
    "  @par     "
    '(p "Parameter:  ")
    'n
    grdoc-comment-start
    "  @pdesc   "
    '(p "Description:  ")
    'n
    grdoc-comment-start
    "  @ptype   " 
    '(p "Type:  ")
    'n
    grdoc-comment-start
    "  @pvalues "
    '(p "Values:  ")
    'n
    grdoc-comment-start
    "  @pcomment \n"
    grdoc-comment-start
    'p 'n
    grdoc-comment-start
    "  @endpar \n"
    )
)

(tempo-define-template
 "grdoc-history-element"
 '(
   grdoc-comment-start
   "  @hdate "
   (current-time-string) 
   " @hauthor "
   (user-full-name)
   'n
   grdoc-comment-start
   "  @hdesc "
   'p 
   )
)

(defun grdoc-file-header ()
  "Insert grdoc comments for a file header.
Insert comments for the file name, date, author, and description." 
  (interactive)
;  (grdoc-set-comment-start)
  (let ((tempo-interactive grdoc-use-interactive))
    (call-interactively 'tempo-template-grdoc-file-header)))

(defun grdoc-header-header ()
  "Insert grdoc comments for a header.
Insert comments for the header name, date, author, and description." 
  (interactive)
;  (grdoc-set-comment-start)
  (let ((tempo-interactive grdoc-use-interactive))
    (call-interactively 'tempo-template-grdoc-header-header)))

(defun grdoc-routine-header ()
  "Insert grdoc comments for a routine.
Insert comments for the routine name, date, author, description,
called functions, called-by, and history.  If `grdoc-use-ai' is t,
will attempt to glean as much information from the source as possible
to fill in the blanks.  If `grdoc-use-interactive' is non-nil, will
prompt for name, description, etc."

  (interactive)
;  (grdoc-set-comment-start) 
  (let ((tempo-interactive grdoc-use-interactive))
    (call-interactively 'tempo-template-grdoc-routine-header)))

(defun grdoc-var-element ()
  "Insert grdoc comments for a variable.
Insert comments for the variable name, description, type,
input/output use and comment.  If `grdoc-use-ai' is t, will attempt to
glean as much information from the source as possible to fill in the
blanks.  If `grdoc-use-interactive' is non-nil, will prompt for
parameter name, description, type, possible values and comment."
  (interactive)
;  (grdoc-set-comment-start) 
  (let ((tempo-interactive grdoc-use-interactive))
    (call-interactively 'tempo-template-grdoc-var-element)))

(defun grdoc-par-element ()
  "Insert grdoc comments for a parameter.
Insert comments for the parameter name, description, type, possible
values and comment.  If `grdoc-use-ai' is t, will attempt to glean as
much information from the source as possible to fill in the blanks.
If `grdoc-use-interactive' is non-nil, will prompt for parameter name,
description, type, possible values and comment."
  (interactive)
;  (grdoc-set-comment-start) 
  (let ((tempo-interactive grdoc-use-interactive))
    (call-interactively 'tempo-template-grdoc-par-element)))

(defun grdoc-history-element ()
  "Insert grdoc comments for a history section."
(interactive)
;  (grdoc-set-comment-start) 
  (let ((tempo-interactive grdoc-use-interactive))
    (call-interactively 'tempo-template-grdoc-history-element)))

(defun grdoc-routine-name ()
  "Return name of next function for grdoc commenting.
If `grdoc-use-ai' is non-nil, look it up in the source.  Otherwise,
return '(l (p \"Routine:  \"))."
  (if (not grdoc-use-ai)
      '(l (p "Routine:  "))
    (cond
     ((eq major-mode 'fortran-mode) (grdoc-fortran-routine-name))
     (t  '(l (p "Routine:  "))))))

(defun grdoc-fortran-routine-name ()
  (save-excursion
    (if (re-search-forward 
	 "\\(subroutine\\|function\\)[ \t]+\\([A-Za-z0-9_]+\\)[ \t]*(\\([A-Za-z0-9_,]*\\)" nil t)
	(match-string 2)
      '(l (p "Routine:  ")))))

(defun grdoc-calls ()
  "Returns functions current function calls"
  (if (not grdoc-use-ai)
      "\n"
    (cond
     ((eq major-mode 'fortran-mode) (grdoc-fortran-calls))
     (t  "\n"))))

(defun grdoc-fortran-calls ()
  (save-excursion
    ;; Find end of function
    (let ((end (save-excursion (re-search-forward "^[ \t]*end[ \t]*$" nil t)))
          (calls))
      (while (re-search-forward "^[ \t]*call[ \t]+\\([A-Za-z0-9_]+\\)[ \t]*(\\([A-Za-z0-9_,]*\\)" end t)
        (setq calls (concat calls " " (match-string 1))))
      (concat calls "\n"))))
      
(defun grdoc-insert-variables ()
  "Returns stuff for tempo to insert for variables"
  (if (not grdoc-use-ai)
      "\n"
    (cond
     ((eq major-mode 'fortran-mode) (grdoc-fortran-insert-variables))
     (t  "\n"))))

(defun grdoc-fortran-insert-variables ()
  "Find fortran variables.  "
  (save-excursion
    (let ((v-list)
    ;; Find end of function
	  (end (save-excursion (re-search-forward "^[ \t]*end[ \t]*$" nil t)))
	  (out))
      (while (re-search-forward "[ \t]*\\(real\\|integer\\|double\\|complex\\|logical\\|character\\)[ \t]+.*\\(\n     [^ ]  .*\\)*" end t)
	(let ((string (match-string 0))
	      (type)
	      (next))
	  (save-excursion 
	    (get-buffer-create " grdoc-rem-cont")
	    (set-buffer " grdoc-rem-cont")
	    (erase-buffer)
	    (insert string)
	    (goto-char (point-min))
	    (while (re-search-forward "\n     [^ ]  " nil t)
	      (replace-match "" nil nil))
	    (setq string (buffer-string)))
	  (string-match "\\(real\\|integer\\|double\\|complex\\|logical\\|character\\)" string)
	  (setq type (substring string (match-beginning 0) (match-end 0))
		next (1+ (match-end 0)))
	  ;; pick up all identifiers
	  (if (equal grdoc-ai-var-style 'multi)
	      (setq v-list (cons (cons type (substring string next)) v-list))
	    ;; Separate the variables
	    (while (string-match "[A-Za-z0-9_]+" string next)
	      (setq next (match-end 0))
	      (setq v-list (cons (cons type (substring string (match-beginning 0))) v-list))))
	  (setq out (list 'l ))))
      (while v-list
	(let ((name (cdar v-list))
	      (type (caar v-list)))
	  (setq out (append out 
			    (list 
			     grdoc-comment-start
			     "  @var    " name 'n
			     grdoc-comment-start
			     "  @vdesc  " (list 'p (concat "Description of " name ":  ")) 'n
			     grdoc-comment-start
			     "  @vtype  " type 'n
			     grdoc-comment-start
			     "  @vio    " (list 'l (list 'grdoc-prompt-vio name)) 'n
			     grdoc-comment-start
			     "  @vcomment \n"
			     grdoc-comment-start
			     'p 'n
			     grdoc-comment-start
			     "  @endvar \n"))
			    v-list (cdr v-list))))
	out)))

(defun grdoc-prompt-vio (&optional name)
  "Prompt for the i/o of the variable. 
Choose in, out, inout, or local, commonblock, or global with completion."
  (if grdoc-use-interactive
      (completing-read (concat nil " IO type:  ")
		       '(("in" 1)
			 ("out" 2)
			 ("inout" 3)
			 ("local" 4)
			 ("commonblock" 5)
			 ("global" 6))
		       nil t)))
	      
(defun grdoc-match-string (num &optional string)
  "Return string of text matched by last search.
NUM specifies which parenthesized expression in the last regexp.
Value is nil if NUMth pair didn't match, or there were less than NUM pairs.
Zero means the entire text matched by the whole regexp or whole string.
STRING should be given if the last search was by `string-match' on STRING."
  (if (match-beginning num)
      (if string
	  (substring string (match-beginning num) (match-end num))
	(buffer-substring (match-beginning num) (match-end num)))))

;; Add grdoc mode to minor-mode-alist
(or (assq 'grdoc-mode minor-mode-alist)
    (setq minor-mode-alist
	  (cons (list 'grdoc-mode 
		      " grdoc")
		minor-mode-alist)))

;; Add the minor-mode keymaps
(or (assq 'grdoc-mode-map minor-mode-map-alist)
    (setq minor-mode-map-alist
	  (cons (cons 'grdoc-mode grdoc-mode-map) minor-mode-map-alist)))

(provide 'grdoc)
;;; GRDOC.EL ends here
